from typing import List, Tuple
import numpy as np
import pandas as pd
from pydantic import BaseModel, Field
from sklearn.model_selection import StratifiedKFold, train_test_split


class DatasetSplitter(BaseModel):
    fold_size: int = Field(default=5, ge=2)
    test_size: float = Field(default=0.2, ge=0.0, le=1.0)
    shuffle: bool = True
    random_state: int = Field(default=42)

    class Config:
        arbitrary_types_allowed = True

    def split(self, X: pd.DataFrame, y: pd.Series) -> Tuple[
            pd.DataFrame, pd.DataFrame, pd.Series, pd.Series, pd.Index, pd.Index, List[np.ndarray], List[np.ndarray]]:
        print(
            f"Splitting dataset into {self.fold_size} folds with shuffle={self.shuffle}...")
        # TODO check it
        if len(y.shape) == 1:
            y = pd.Series(np.expand_dims(y.to_numpy(), axis=1).flatten())
        stacked = np.vstack(tuple(y.to_numpy()))
        stratify = np.argmax(stacked, axis=1)
        X_train, X_test, y_train, y_test = train_test_split(
            X, y, shuffle=self.shuffle, test_size=self.test_size, stratify=stratify)

        k_fold = StratifiedKFold(
            n_splits=self.fold_size, shuffle=self.shuffle, random_state=self.random_state)
        folds = k_fold.split(X_train, np.argmax(
            np.vstack(y_train.to_numpy()), axis=1))
        train_idx_arr = []
        val_idx_arr = []
        for i, (train_index, val_index) in enumerate(folds):
            train_idx_arr.append(train_index)
            val_idx_arr.append(val_index)

        return X_train, X_test, y_train, y_test, X_train.index, X_test.index, train_idx_arr, val_idx_arr



import numpy as np

def adjust_array_dims(arr, final_ndim=2):
    # Add axes if array has fewer dimensions than final_ndim
    while arr.ndim < final_ndim:
        arr = arr[:, np.newaxis]  # Add a new axis
        
    # Drop axes if array has more dimensions than final_ndim
    while arr.ndim > final_ndim:
        arr = np.squeeze(arr, axis=-1)  # Remove the last axis
    
    return arr



# # Example usage
# arr_1d = np.array([1, 2, 3, 4, 5])

# # Convert to a 3D array (iteratively adds axes)
# arr_3d = adjust_array_dims(arr_1d, final_ndim=3)
# print(arr_3d)
# print("Shape of arr_3d:", arr_3d.shape)

# # Convert to a 2D array (iteratively drops axes)
# arr_2d = adjust_array_dims(arr_3d, final_ndim=2)
# print(arr_2d)
# print("Shape of arr_2d:", arr_2d.shape)


from .utils import clear_directory,create_folder_if_not_exists, utc_time_as_string,utc_time_as_string_simple_format, compress_and_save_data
from .zip_helper import ZipHelper
from .py7zr_helper import Py7ZipHelper
from .enums import UMLSCodeTypes, DrugBankTextDataTypes
from .package_helper import get_import
from .kaggle import create_kaggle_dataset
from .categorical_data_encoding_checker import to_one_hot_encode,is_one_hot_encoded, is_binary_encoded, is_binary_vector,is_label_encoded
from .numpy_utils import adjust_array_dims



from sqlite3 import Error
import sqlite3
import pandas as pd
import numpy as np


def create_connection(db_file=r"./event.db"):
    """ create a database connection to the SQLite database
        specified by db_file
    :param db_file: database file
    :return: Connection object or None
    """
    conn = None
    try:
        conn = sqlite3.connect(db_file)
    except Error as e:
        print(e)

    return conn


# def select_all_drugs(conn):
#     cur = conn.cursor()
#     cur.execute(
#         '''select "index", id, name, target, enzyme, pathway, smile from drug''')
#     rows = cur.fetchall()
#     return rows


# def select_all_drugs_as_dataframe(conn):
#     headers = ['index','id', 'name', 'target', 'enzyme', 'pathway', 'smile']
#     rows = select_all_drugs(conn)
#     df = pd.DataFrame(columns=headers, data=rows)
#     df['enzyme'] = df['enzyme'].apply(lambda x: x.split('|'))
#     df['target'] = df['target'].apply(lambda x: x.split('|'))
#     df['pathway'] = df['pathway'].apply(lambda x: x.split('|'))
#     df['smile'] = df['smile'].apply(lambda x: x.split('|'))
#     return df


# def select_all_events(conn):
#     """
#     Query all rows in the event table
#     :param conn: the Connection object
#     :return:
#     """
#     cur = conn.cursor()
#     cur.execute("select * from event")

#     rows = cur.fetchall()
#     return rows


# def select_all_events_as_dataframe(conn):
#     headers = ["index", "id1", "name1", "id2", "name2", "event_category"]
#     rows = select_all_events(conn)
#     return pd.DataFrame(columns=headers, data=rows)


# def select_events_with_category(conn):
#     sql = '''select id1, name1, id2, name2, mechanism || ' ' ||action from event ev
#             join extraction ex
#             on ev.name1 = ex.drugA and ev.name2 = ex.drugB
#             union 
#             select id1, name1, id2, name2, mechanism || ' ' ||action from event ev
#             join extraction ex
#             on ev.name1 = ex.drugB and ev.name2 = ex.drugA
#             '''
#     cur = conn.cursor()
#     cur.execute(sql)

#     rows = cur.fetchall()

#     headers = ['id1', 'name1', 'id2', 'name2', 'event_category']
#     return pd.DataFrame(columns=headers, data=rows)


# def select_all_interactions_tuple_as_dataframe(conn):
#     cur = conn.cursor()
#     cur.execute("select id1, id2 from event")
#     rows = cur.fetchall()
#     headers = ['id1', 'id2']

#     return pd.DataFrame(columns=headers, data=rows)


# def select_ddi_pairs(conn):
#     cur = conn.cursor()
#     cur.execute('''
#         select d1.[index] as Drug1Index, d2.[index] as Drug2Index, 1 from event e 
#         join drug d1 on e.id1 = d1.id 
#         join drug d2 on e.id2 = d2.id 
#     ''')
#     rows = cur.fetchall()
#     return rows


# def select_ddi_pairs_as_dataframe(conn):
#     headers = ["Drug1Index", "Drug2Index", "Interaction"]
#     rows = select_ddi_pairs(conn)
#     return pd.DataFrame(columns=headers, data=rows)


# def get_interactions(conn):
#     cur = conn.cursor()
#     cur.execute('''
#         select
#             drug_1_id,
#             drug_1,
#             drug_2_id,
#             drug_2,
#             mechanism_action, 
#             interaction, 
#             masked_interaction
#         from _Interactions
#                 ''')

#     rows = cur.fetchall()

#     headers = ['id1', 'name1', 'id2', 'name2',
#                'event_category', 'interaction', 'masked_interaction']
#     df = pd.DataFrame(columns=headers, data=rows)
#     return df


# def get_extended_version(conn):
#     cur = conn.cursor()
#     cur.execute('''
#         select 
#         _Drugs."index",
#         drugbank_id,
#         _Drugs.name,
#         description,
#         synthesis_reference,
#         indication,
#         pharmacodynamics,
#         mechanism_of_action,
#         toxicity,
#         metabolism,
#         absorption,
#         half_life,
#         protein_binding,
#         route_of_elimination,
#         volume_of_distribution,
#         clearance,
#         smiles,
#         smiles_morgan_fingerprint,
#         enzymes_polypeptides,
#         targets_polypeptides
                
#         from drug 
#         join _Drugs on drug.id = _Drugs.drugbank_id  
#         where 
#                 targets_polypeptides is not null and 
#                 enzymes_polypeptides is not null and 
#                 smiles_morgan_fingerprint is not null
# ''')
#     # pathway is absent

#     rows = cur.fetchall()
#     headers = ['index', 'id', 'name', 'description', 'synthesis_reference', 'indication', 'pharmacodynamics', 'mechanism_of_action', 'toxicity', 'metabolism', 'absorption', 'half_life',
#                'protein_binding', 'route_of_elimination', 'volume_of_distribution', 'clearance', 'smiles_notation', 'smile', 'enzyme', 'target']
#     df = pd.DataFrame(columns=headers, data=rows)
#     df['smile'] = df['smile'].apply(lambda x:
#                                     np.fromstring(
#                                         x.replace(
#                                             '\n', '')
#                                         .replace('[', '')
#                                         .replace(']', '')
#                                         .replace('  ', ' '), sep=','))
#     df['enzyme'] = df['enzyme'].apply(
#         lambda x: x.split('|'))
#     df['target'] = df['target'].apply(
#         lambda x: x.split('|'))
#     return df


# SELECT
#  CASE
#  WHEN masked_interaction like '%'+drug_1+'%' THEN drug_1
#  WHEN masked_interaction like '%'+drug_2+'%' THEN drug_2
#  Else drug_2
#  END AS Absent,

# drug_1, drug_2,
# masked_interaction

# from _Interactions
# where LENGTH(masked_interaction) = LENGTH(REPLACE(masked_interaction, 'DRUG', ''))
# or LENGTH(masked_interaction) = LENGTH(REPLACE(masked_interaction, 'DRUG', '')) + 4

# if __name__ == "__main__":
#     conn = create_connection(r"./event-extended.db")
#     extended_version_df = get_extended_version(conn)

#     df = select_all_events_as_dataframe(conn)
#     print(df.head())

#     events_with_category_df = select_events_with_category(conn)
#     print(events_with_category_df.head())

#     u = events_with_category_df['event_category'].unique()
#     print(len(u))



from typing import List
import nltk
from nltk import sent_tokenize
from langchain_text_splitters.base import TextSplitter

nltk.download('punkt')

class SentenceSplitter(TextSplitter):
    def split_text(self, text: str) -> List[str]:
        return sent_tokenize(text)



import json
import sys

def minify(folder, file_name):
    "Minify JSON"
    file_data = open(folder+'/'+file_name, "r", 1).read() # store file info in variable
    json_data = json.loads(file_data) # store in json structure
    json_string = json.dumps(json_data, separators=(',', ":")) # Compact JSON structure
    file_name = str(file_name).replace(".json", "") # remove .json from end of file_name string
    new_file_name = folder+"/{0}_minify.json".format(file_name)
    open(new_file_name, "w+", 1).write(json_string) # open and write json_string to file


from abc import ABC, abstractmethod
import itertools


class CombinationStrategy(ABC):
    @abstractmethod
    def generate(self) -> list:
        pass


class CustomCombinationStrategy(CombinationStrategy):
    def __init__(self, **kwargs_combination_params):
        # kwargs fonksiyona da alınabilir
        self.group_1 = kwargs_combination_params.get("group_1", None)
        self.group_2 = kwargs_combination_params.get("group_2", None)

    def generate(self):
        # Handle edge cases
        if not self.group_1 or not self.group_2:
            raise ValueError(
                f"Parameters of combination strategy could not be empty.")
            # return []  # Return an empty list if either group is empty
        # combinations = []
        # for j in self.group2:
        #     extended_item_group_1 = self.group_1.copy()
        #     extended_item_group_1.append(j)
        #     for i in range(2, len(extended_item_group_1) + 1):
        #         combinations.extend(list(itertools.combinations(extended_item_group_1, i))) #all
        # combinations = list(set(combinations))

        combinations = set()  # Use a set to avoid duplicates directly
        for j in self.group_2:
            extended_item_group_1 = self.group_1.copy()
            extended_item_group_1.append(j)
            # Generate combinations of all lengths from 2 to len(group_1 + 1)
            for i in range(2, len(extended_item_group_1) + 1):
                combinations.update(itertools.combinations(
                    extended_item_group_1, i))  # Add combinations

        # Convert set back to list (if needed) and return
        return list(combinations)



from typing import Any, Callable
from ddi_fw.ml.model_wrapper import ModelWrapper
import tensorflow as tf
from tensorflow import keras
# from keras.callbacks import EarlyStopping, ModelCheckpoint
from tensorflow.keras.callbacks import EarlyStopping, ModelCheckpoint, Callback
from sklearn.model_selection import train_test_split, KFold, StratifiedKFold
import numpy as np
from tensorflow.keras import Model
from ddi_fw.ml.evaluation_helper import Metrics, evaluate

# import tf2onnx
# import onnx

from ddi_fw.ml.tracking_service import TrackingService
import ddi_fw.utils as utils
import os


def convert_to_categorical(arr, num_classes):
    """
    This function takes an array of labels and converts them to one-hot encoding 
    if they are not binary-encoded. If the array is already in a 
    compatible format, it returns the original array.

    Parameters:
    - arr: numpy array with label data (could be binary-encoded or label-encoded)
    - num_classes: number of classes to be used in one-hot encoding

    Returns:
    - The one-hot encoded array if the original array was binary or label encoded
    - The original array if it doesn't require any conversion
    """

    try:
        # First, check if the array is binary-encoded
        if not utils.is_binary_encoded(arr):
            # If the arr labels are binary-encoded, convert them to one-hot encoding
            return tf.keras.utils.to_categorical(np.argmax(arr, axis=1), num_classes=num_classes)
        else:
            print("No conversion needed, returning original array.")
            return arr
    except Exception as e:
        # If binary encoding check raises an error, print it and continue to label encoding check
        print(f"Error while checking binary encoding: {e}")

    try:
        # Check if the array is label-encoded
        if utils.is_label_encoded(arr):
            # If the arr labels are label-encoded, convert them to one-hot encoding
            return tf.keras.utils.to_categorical(arr, num_classes=num_classes)
    except Exception as e:
        # If label encoding check raises an error, print it
        print(f"Error while checking label encoding: {e}")
        # If the arr labels don't match any of the known encodings, raise an error
        raise ValueError("Unknown label encoding format.")

    # If no conversion was needed, return the original array

    return arr


class TFModelWrapper(ModelWrapper):

    def __init__(self, date, descriptor, model_func, tracking_service: TrackingService | None = None, **kwargs):
        super().__init__(date, descriptor, model_func, **kwargs)
        self.batch_size = kwargs.get('batch_size', 128)
        self.epochs = kwargs.get('epochs', 100)
        self.tracking_service = tracking_service

    # TODO think different settings for num_classes
    def fit_model(self, X_train, y_train, X_valid, y_valid):
        self.kwargs['input_shape'] = self.train_data.shape
        self.num_classes = len(np.unique(y_train, axis=0))
        self.kwargs['num_classes'] = self.num_classes
        model = self.model_func(**self.kwargs)
        checkpoint = ModelCheckpoint(
            filepath=f'{self.descriptor}_validation.weights.h5',
            monitor='val_loss',
            save_best_only=True,
            save_weights_only=True,
            verbose=1,
            mode='min'
        )
        early_stopping = EarlyStopping(
            monitor='val_loss', patience=10, mode='auto')
        if self.tracking_service:
            custom_callback = CustomCallback(self.tracking_service)
            callbacks=[early_stopping, checkpoint, custom_callback]
        else:
            callbacks=[early_stopping, checkpoint]
        train_dataset = tf.data.Dataset.from_tensor_slices((X_train, y_train))
        train_dataset = train_dataset.batch(batch_size=self.batch_size)

        if X_valid is not None and y_valid is not None:
            val_dataset = tf.data.Dataset.from_tensor_slices(
                (X_valid, y_valid))
            val_dataset = val_dataset.batch(batch_size=self.batch_size)
        else:
            val_dataset = None

        history = model.fit(
            train_dataset,
            epochs=self.epochs,
            validation_data=val_dataset,
            callbacks=callbacks
        )
        
            # Check if early stopping was applied
        if early_stopping.stopped_epoch > 0:
            print(f"Early stopping was applied at epoch {early_stopping.stopped_epoch}.")
        else:
            print("Early stopping was not applied.")
        if self.tracking_service:
            self.tracking_service.log_param("early_stopping_applied", early_stopping.stopped_epoch > 0)
            self.tracking_service.log_param("early_stopping_epoch", early_stopping.stopped_epoch)
        # ex
        # history = model.fit(
        #     X_train, y_train,
        #     batch_size=self.batch_size,
        #     epochs=self.epochs,
        #     validation_data=(X_valid, y_valid),
        #     callbacks=[early_stopping, checkpoint, custom_callback]
        # )

        if os.path.exists(f'{self.descriptor}_validation.weights.h5'):
            os.remove(f'{self.descriptor}_validation.weights.h5')

        return checkpoint.model, checkpoint

    def fit(self):
        print(self.train_data.shape)
        models = {}
        models_val_acc = {}
        if self.train_idx_arr and self.val_idx_arr:
            for i, (train_idx, val_idx) in enumerate(zip(self.train_idx_arr, self.val_idx_arr)):
                print(f"Validation {i}")
                X_train_cv = self.train_data[train_idx]
                y_train_cv = self.train_label[train_idx]
                X_valid_cv = self.train_data[val_idx]
                y_valid_cv = self.train_label[val_idx]

                def fit_model_cv_func():
                    model, checkpoint = self.fit_model(
                        X_train_cv, y_train_cv, X_valid_cv, y_valid_cv)
                    return model, checkpoint

                if self.tracking_service:
                    model, checkpoint = self.tracking_service.run(
                        run_name=f'Validation {i}', description='CV models', nested_run=True, func=fit_model_cv_func)
                    # with mlflow.start_run(run_name=f'Validation {i}', description='CV models', nested=True) as cv_fit:

                    #     model, checkpoint = self.fit_model(
                    #         X_train_cv, y_train_cv, X_valid_cv, y_valid_cv)
                    #     models[f'{self.descriptor}_validation_{i}'] = model
                    #     models_val_acc[f'{self.descriptor}_validation_{i}'] = checkpoint.best
                else:
                    model, checkpoint = fit_model_cv_func()
                    # model, checkpoint = self.fit_model(
                    #     X_train_cv, y_train_cv, X_valid_cv, y_valid_cv)
                models[f'{self.descriptor}_validation_{i}'] = model
                models_val_acc[f'{self.descriptor}_validation_{i}'] = checkpoint.best
        else:
            def fit_model_func():
                model, checkpoint = self.fit_model(
                    self.train_data, self.train_label, None, None)
                return model, checkpoint

            if self.tracking_service:
                model, checkpoint = self.tracking_service.run(
                    run_name=f'Training', description='Training', nested_run=True, func=fit_model_func)
                # with mlflow.start_run(run_name=f'Training', description='Training', nested=True) as cv_fit:
                #     model, checkpoint = self.fit_model(
                #         self.train_data, self.train_label, None, None)
                #     models[self.descriptor] = model
                #     models_val_acc[self.descriptor] = checkpoint.best
            else:
                model, checkpoint = fit_model_func()
                # models[self.descriptor] = model
                # models_val_acc[self.descriptor] = checkpoint.best
            models[self.descriptor] = model
            models_val_acc[self.descriptor] = checkpoint.best
        if models_val_acc == {}:
            return model, None
        best_model_key = max(models_val_acc, key=lambda k: models_val_acc[k])
        print("best model key: ", best_model_key)
        # best_model_key = max(models_val_acc, key=models_val_acc.get)
        best_model = models[best_model_key]
        return best_model, best_model_key

    # https://github.com/mlflow/mlflow/blob/master/examples/tensorflow/train.py

    def predict(self):
        test_dataset = tf.data.Dataset.from_tensor_slices(
            (self.test_data, self.test_label))
        test_dataset = test_dataset.batch(batch_size=1)
        # pred = self.best_model.predict(self.test_data)
        pred = self.best_model.predict(test_dataset)
        return pred

    # def fit_and_evaluate(self, print_detail=False) -> tuple[dict[str, Any], Metrics, Any]:
    #     if self.use_mlflow:
    #         with mlflow.start_run(run_name=self.descriptor, description="***", nested=True) as run:
    #             best_model, best_model_key = self.fit()
    #             self.best_model: Model = best_model
    #             pred = self.predict()
    #             actual = self.test_label
    #             # if not utils.is_binary_encoded(pred):
    #             #     pred = tf.keras.utils.to_categorical(np.argmax(pred,axis=1), num_classes=self.num_classes)
    #             pred_as_cat = convert_to_categorical(pred, self.num_classes)
    #             actual_as_cat = convert_to_categorical(
    #                 actual, self.num_classes)

    #             logs, metrics = evaluate(
    #                 actual=actual_as_cat, pred=pred_as_cat, info=self.descriptor, print_detail=print_detail)
    #             metrics.format_float()
    #             mlflow.log_metrics(logs)
    #             mlflow.log_param('best_cv', best_model_key)
    #             utils.compress_and_save_data(
    #                 metrics.__dict__, run.info.artifact_uri, f'{self.date}_metrics.gzip')
    #             mlflow.log_artifact(
    #                 f'{run.info.artifact_uri}/{self.date}_metrics.gzip')

    #             return logs, metrics, pred
    #     else:
    #         best_model, best_model_key = self.fit()
    #         self.best_model = best_model
    #         pred = self.predict()
    #         actual = self.test_label

    #         pred_as_cat = convert_to_categorical(pred, self.num_classes)
    #         actual_as_cat = convert_to_categorical(actual, self.num_classes)
    #         logs, metrics = evaluate(
    #             actual=actual_as_cat, pred=pred_as_cat, info=self.descriptor)
    #         metrics.format_float()
    #         return logs, metrics, pred

    def fit_and_evaluate(self, print_detail=False) -> tuple[dict[str, Any], Metrics, Any]:
        """
        Fit the model, evaluate it, and log results using the tracking service.

        Args:
            print_detail (bool): Whether to print detailed evaluation logs.

        Returns:
            tuple: A tuple containing logs, metrics, and predictions.
        """
        self.best_model: Model = None
        def evaluate_and_log(artifact_uri=None):
            # Fit the model
            best_model, best_model_key = self.fit()
            self.best_model = best_model

            # Make predictions
            pred = self.predict()
            actual = self.test_label

            # Convert predictions and actual labels to categorical format
            pred_as_cat = convert_to_categorical(pred, self.num_classes)
            actual_as_cat = convert_to_categorical(actual, self.num_classes)

            # Evaluate the model
            logs, metrics = evaluate(
                actual=actual_as_cat, pred=pred_as_cat, info=self.descriptor, print_detail=print_detail
            )
            metrics.format_float()

            if self.tracking_service:
                # Log metrics and parameters
                self.tracking_service.log_metrics(logs)
                self.tracking_service.log_param('best_cv', best_model_key)

                # Save metrics to the artifact URI if provided
                if artifact_uri:
                    utils.compress_and_save_data(
                        metrics.__dict__, artifact_uri, f'{self.date}_metrics.gzip'
                    )
                    self.tracking_service.log_artifact(
                        f'{artifact_uri}/{self.date}_metrics.gzip'
                    )

            return logs, metrics, pred

        # Use the tracking service to run the evaluation
        if self.tracking_service:
            return self.tracking_service.run(
                run_name=self.descriptor,
                description="Fit and evaluate the model",
                nested_run=True,
                func=evaluate_and_log
            )
        else:
            # If no tracking service is provided, run the evaluation directly
            return evaluate_and_log()


class CustomCallback(Callback):
    """
    Custom Keras callback for logging training metrics and model summary to MLflow.
    """
    def __init__(self, tracking_service: TrackingService):
        super().__init__()
        self.tracking_service = tracking_service

    # def _mlflow_log(self, func: Callable):
    #     if self.use_mlflow:
    #         func()

    def on_train_begin(self, logs=None):
        if logs is None:
            logs = {}
        if not isinstance(self.model, Model):
            raise TypeError("self.model must be an instance of Model")

        keys = list(logs.keys())

        self.tracking_service.log_param("train_begin_keys", keys)
        # self._mlflow_log(lambda: mlflow.log_param("train_begin_keys", keys))

        # config = self.model.optimizer.get_config()
        config = self.model.get_config()
        for attribute in config:
            self.tracking_service.log_param(
                "opt_" + attribute, config[attribute])
            # self._mlflow_log(lambda: mlflow.log_param(
            #     "opt_" + attribute, config[attribute]))

        sum_list = []
        self.model.summary(print_fn=sum_list.append)
        summary = "\n".join(sum_list)
        self.tracking_service.log_text(
            summary, file_name="model_summary.txt")
        # self._mlflow_log(lambda: mlflow.log_text(
        #     summary, artifact_file="model_summary.txt"))

    def on_train_end(self, logs=None):
        if logs is None:
            logs = {}
        print(logs)
        self.tracking_service.log_metrics(logs)
        # self._mlflow_log(lambda: mlflow.log_metrics(logs))

    def on_epoch_begin(self, epoch, logs=None):
        if logs is None:
            logs = {}
        keys = list(logs.keys())

    def on_epoch_end(self, epoch, logs=None):
        if logs is None:
            logs = {}
        keys = list(logs.keys())

    def on_test_begin(self, logs=None):
        if logs is None:
            logs = {}
        keys = list(logs.keys())

    def on_test_end(self, logs=None):
        if logs is None:
            logs = {}
        self.tracking_service.log_metrics(logs)
        # self._mlflow_log(lambda: mlflow.log_metrics(logs))
        print(logs)

    def on_predict_begin(self, logs=None):
        if logs is None:
            logs = {}
        keys = list(logs.keys())

    def on_predict_end(self, logs=None):
        if logs is None:
            logs = {}
        keys = list(logs.keys())
        self.tracking_service.log_metrics(logs)
        # self._mlflow_log(lambda: mlflow.log_metrics(logs))

    def on_train_batch_begin(self, batch, logs=None):
        if logs is None:
            logs = {}
        keys = list(logs.keys())

    def on_train_batch_end(self, batch, logs=None):
        if logs is None:
            logs = {}
        keys = list(logs.keys())

    def on_test_batch_begin(self, batch, logs=None):
        if logs is None:
            logs = {}
        keys = list(logs.keys())

    def on_test_batch_end(self, batch, logs=None):
        if logs is None:
            logs = {}
        keys = list(logs.keys())

    def on_predict_batch_begin(self, batch, logs=None):
        if logs is None:
            logs = {}
        keys = list(logs.keys())

    def on_predict_batch_end(self, batch, logs=None):
        if logs is None:
            logs = {}
        keys = list(logs.keys())
